print("import: 'sqlalchemy_utils'")
import sqlalchemy_utils

print("import: 'sqlalchemy_utils.functions'")
import sqlalchemy_utils.functions

print("import: 'sqlalchemy_utils.primitives'")
import sqlalchemy_utils.primitives

print("import: 'sqlalchemy_utils.relationships'")
import sqlalchemy_utils.relationships

print("import: 'sqlalchemy_utils.types'")
import sqlalchemy_utils.types

